@@ -13,11 +13,29 @@
 #define POLLY_ISL_EXPR_BUILDER_H
 
 #include "polly/CodeGen/IRBuilder.h"
+#include "polly/Support/ScopHelper.h"
+
 #include "llvm/ADT/MapVector.h"
 #include "isl/ast.h"
 
 namespace llvm {
-class SCEVExpander;
+class DataLayout;
+class ScalarEvolution;
+}
+
+struct isl_id;
+
+namespace llvm {
+// Provide PointerLikeTypeTraits for isl_id.
+template <> class PointerLikeTypeTraits<isl_id *> {
+
+public:
+  static inline const void *getAsVoidPointer(isl_id *P) { return (void *)P; }
+  static inline const Region *getFromVoidPointer(void *P) {
+    return (Region *)P;
+  }
+  enum { NumLowBitsAvailable = 0 };
+};
 }
 
 namespace polly {
@@ -79,7 +97,7 @@ namespace polly {
 class IslExprBuilder {
 public:
   /// @brief A map from isl_ids to llvm::Values.
-  typedef llvm::MapVector<isl_id *, llvm::Value *> IDToValueTy;
+  typedef llvm::MapVector<isl_id *, llvm::AssertingVH<llvm::Value>> IDToValueTy;
 
   /// @brief Construct an IslExprBuilder.
   ///
@@ -91,13 +109,12 @@ public:
   ///                  variables (identified by an isl_id). The IDTOValue map
   ///                  specifies the LLVM-IR Values that correspond to these
   ///                  parameters and variables.
-  /// @param Expander  A SCEVExpander to create the indices for multi
-  ///                  dimensional accesses.
-  IslExprBuilder(PollyIRBuilder &Builder, IDToValueTy &IDToValue,
-                 llvm::SCEVExpander &Expander, llvm::DominatorTree &DT,
+  IslExprBuilder(Scop &S, PollyIRBuilder &Builder, IDToValueTy &IDToValue,
+                 ValueMapT &GlobalMap, const llvm::DataLayout &DL,
+                 llvm::ScalarEvolution &SE, llvm::DominatorTree &DT,
                  llvm::LoopInfo &LI)
-      : Builder(Builder), IDToValue(IDToValue), Expander(Expander), DT(DT),
-        LI(LI) {}
+      : S(S), Builder(Builder), IDToValue(IDToValue), GlobalMap(GlobalMap),
+        DL(DL), SE(SE), DT(DT), LI(LI) {}
 
   /// @brief Create LLVM-IR for an isl_ast_expr[ession].
   ///
@@ -124,12 +141,14 @@ public:
   llvm::IntegerType *getType(__isl_keep isl_ast_expr *Expr);
 
 private:
+  Scop &S;
+
   PollyIRBuilder &Builder;
   IDToValueTy &IDToValue;
+  ValueMapT &GlobalMap;
 
-  /// @brief A SCEVExpander to translate dimension sizes to llvm values.
-  llvm::SCEVExpander &Expander;
-
+  const llvm::DataLayout &DL;
+  llvm::ScalarEvolution &SE;
   llvm::DominatorTree &DT;
   llvm::LoopInfo &LI;
 
