@@ -43,6 +43,9 @@ class Region;
 
 namespace polly {
 
+/// @brief Set the begin and end source location for the given region @p R.
+void getDebugLocations(const Region *R, DebugLoc &Begin, DebugLoc &End);
+
 class RejectLog;
 /// @brief Emit optimization remarks about the rejected regions to the user.
 ///
@@ -52,17 +55,11 @@ class RejectLog;
 /// @param Log The error log containing all messages being emitted as remark.
 void emitRejectionRemarks(const llvm::Function &F, const RejectLog &Log);
 
-/// @brief Emit diagnostic remarks for a valid Scop
-///
-/// @param F The function we emit remarks for
-/// @param R The region that marks a valid Scop
-void emitValidRemarks(const llvm::Function &F, const Region *R);
-
 // Discriminator for LLVM-style RTTI (dyn_cast<> et al.)
 enum RejectReasonKind {
   // CFG Category
   rrkCFG,
-  rrkNonBranchTerminator,
+  rrkInvalidTerminator,
   rrkCondition,
   rrkLastCFG,
 
@@ -80,27 +77,18 @@ enum RejectReasonKind {
   rrkDifferentElementSize,
   rrkLastAffFunc,
 
-  // IndVar
-  rrkIndVar,
-  rrkPhiNodeRefInRegion,
-  rrkLastIndVar,
-
-  rrkIndEdge,
-
   rrkLoopBound,
 
   rrkFuncCall,
+  rrkNonSimpleMemoryAccess,
 
   rrkAlias,
 
-  rrkSimpleLoop,
-
   // Other
   rrkOther,
   rrkIntToPtr,
   rrkAlloca,
   rrkUnknownInst,
-  rrkPHIinExit,
   rrkEntry,
   rrkUnprofitable,
   rrkLastOther
@@ -238,36 +226,13 @@ public:
 };
 
 //===----------------------------------------------------------------------===//
-/// @brief Captures a non-branch terminator within a Scop candidate.
-class ReportNonBranchTerminator : public ReportCFG {
+/// @brief Captures bad terminator within a Scop candidate.
+class ReportInvalidTerminator : public ReportCFG {
   BasicBlock *BB;
 
 public:
-  ReportNonBranchTerminator(BasicBlock *BB)
-      : ReportCFG(rrkNonBranchTerminator), BB(BB) {}
-
-  /// @name LLVM-RTTI interface
-  //@{
-  static bool classof(const RejectReason *RR);
-  //@}
-
-  /// @name RejectReason interface
-  //@{
-  virtual std::string getMessage() const override;
-  virtual const DebugLoc &getDebugLoc() const override;
-  //@}
-};
-
-//===----------------------------------------------------------------------===//
-/// @brief Captures a not well-structured condition within the CFG.
-class ReportCondition : public ReportCFG {
-  //===--------------------------------------------------------------------===//
-
-  // The BasicBlock we found the broken condition in.
-  BasicBlock *BB;
-
-public:
-  ReportCondition(BasicBlock *BB) : ReportCFG(rrkCondition), BB(BB) {}
+  ReportInvalidTerminator(BasicBlock *BB)
+      : ReportCFG(rrkInvalidTerminator), BB(BB) {}
 
   /// @name LLVM-RTTI interface
   //@{
@@ -555,62 +520,6 @@ public:
   //@}
 };
 
-//===----------------------------------------------------------------------===//
-/// @brief Base class for reject reasons related to induction variables.
-///
-//  ReportIndVar reject reasons are generated when the ScopDetection finds
-/// errors in the induction variable(s) of the Scop candidate.
-class ReportIndVar : public RejectReason {
-  //===--------------------------------------------------------------------===//
-public:
-  ReportIndVar(const RejectReasonKind K);
-};
-
-//===----------------------------------------------------------------------===//
-/// @brief Captures a phi node that refers to SSA names in the current region.
-class ReportPhiNodeRefInRegion : public ReportIndVar {
-  //===--------------------------------------------------------------------===//
-
-  // The offending instruction.
-  Instruction *Inst;
-
-public:
-  ReportPhiNodeRefInRegion(Instruction *Inst);
-
-  /// @name LLVM-RTTI interface
-  //@{
-  static bool classof(const RejectReason *RR);
-  //@}
-
-  /// @name RejectReason interface
-  //@{
-  virtual std::string getMessage() const override;
-  virtual const DebugLoc &getDebugLoc() const override;
-  //@}
-};
-
-//===----------------------------------------------------------------------===//
-/// @brief Captures a region with invalid entering edges.
-class ReportIndEdge : public RejectReason {
-  //===--------------------------------------------------------------------===//
-
-  BasicBlock *BB;
-
-public:
-  ReportIndEdge(BasicBlock *BB);
-
-  /// @name LLVM-RTTI interface
-  //@{
-  static bool classof(const RejectReason *RR);
-  //@}
-
-  /// @name RejectReason interface
-  //@{
-  virtual std::string getMessage() const override;
-  virtual const DebugLoc &getDebugLoc() const override;
-  //@}
-};
-
 //===----------------------------------------------------------------------===//
 /// @brief Captures errors with non affine loop bounds.
 class ReportLoopBound : public RejectReason {
@@ -705,24 +614,6 @@ public:
   //@}
 };
 
-//===----------------------------------------------------------------------===//
-/// @brief Captures errors with non simplified loops.
-class ReportSimpleLoop : public RejectReason {
-  //===--------------------------------------------------------------------===//
-public:
-  ReportSimpleLoop();
-
-  /// @name LLVM-RTTI interface
-  //@{
-  static bool classof(const RejectReason *RR);
-  //@}
-
-  /// @name RejectReason interface
-  //@{
-  virtual std::string getMessage() const override;
-  //@}
-};
-
 //===----------------------------------------------------------------------===//
 /// @brief Base class for otherwise ungrouped reject reasons.
 class ReportOther : public RejectReason {
@@ -806,27 +697,6 @@ public:
   //@}
 };
 
-//===----------------------------------------------------------------------===//
-/// @brief Captures errors with phi nodes in exit BBs.
-class ReportPHIinExit : public ReportOther {
-  //===--------------------------------------------------------------------===//
-  Instruction *Inst;
-
-public:
-  ReportPHIinExit(Instruction *Inst);
-
-  /// @name LLVM-RTTI interface
-  //@{
-  static bool classof(const RejectReason *RR);
-  //@}
-
-  /// @name RejectReason interface
-  //@{
-  virtual std::string getMessage() const override;
-  virtual const DebugLoc &getDebugLoc() const override;
-  //@}
-};
-
 //===----------------------------------------------------------------------===//
 /// @brief Captures errors with regions containing the function entry block.
 class ReportEntry : public ReportOther {
@@ -870,6 +740,30 @@ public:
   //@}
 };
 
+//===----------------------------------------------------------------------===//
+/// @brief Captures errors with non-simple memory accesses.
+class ReportNonSimpleMemoryAccess : public ReportOther {
+  //===--------------------------------------------------------------------===//
+
+  // The offending call instruction.
+  Instruction *Inst;
+
+public:
+  ReportNonSimpleMemoryAccess(Instruction *Inst);
+
+  /// @name LLVM-RTTI interface
+  //@{
+  static bool classof(const RejectReason *RR);
+  //@}
+
+  /// @name RejectReason interface
+  //@{
+  virtual std::string getMessage() const override;
+  virtual const DebugLoc &getDebugLoc() const override;
+  virtual std::string getEndUserMessage() const override;
+  //@}
+};
+
 } // namespace polly
 
 #endif // POLLY_SCOP_DETECTION_DIAGNOSTIC_H
