@@ -40,13 +40,10 @@
   STATISTIC(Bad##NAME##ForScop, "Number of bad regions for Scop: " DESC)
 
 BADSCOP_STAT(CFG, "CFG too complex");
-BADSCOP_STAT(IndVar, "Non canonical induction variable in loop");
-BADSCOP_STAT(IndEdge, "Found invalid region entering edges");
 BADSCOP_STAT(LoopBound, "Loop bounds can not be computed");
 BADSCOP_STAT(FuncCall, "Function call with side effects appeared");
 BADSCOP_STAT(AffFunc, "Expression not affine");
 BADSCOP_STAT(Alias, "Found base address alias");
-BADSCOP_STAT(SimpleLoop, "Loop not in -loop-simplify form");
 BADSCOP_STAT(Other, "Others");
 
 namespace polly {
@@ -70,7 +67,7 @@ static bool operator<(const llvm::DebugLoc &LHS, const llvm::DebugLoc &RHS) {
 }
 
 namespace polly {
-static void getDebugLocations(const Region *R, DebugLoc &Begin, DebugLoc &End) {
+void getDebugLocations(const Region *R, DebugLoc &Begin, DebugLoc &End) {
   for (const BasicBlock *BB : R->blocks())
     for (const Instruction &Inst : *BB) {
       DebugLoc DL = Inst.getDebugLoc();
@@ -104,17 +101,6 @@ void emitRejectionRemarks(const llvm::Function &F, const RejectLog &Log) {
                                "Invalid Scop candidate ends here.");
 }
 
-void emitValidRemarks(const llvm::Function &F, const Region *R) {
-  LLVMContext &Ctx = F.getContext();
-
-  DebugLoc Begin, End;
-  getDebugLocations(R, Begin, End);
-
-  emitOptimizationRemark(Ctx, DEBUG_TYPE, F, Begin,
-                         "A valid Scop begins here.");
-  emitOptimizationRemark(Ctx, DEBUG_TYPE, F, End, "A valid Scop ends here.");
-}
-
 //===----------------------------------------------------------------------===//
 // RejectReason.
 const DebugLoc RejectReason::Unknown = DebugLoc();
@@ -143,33 +129,18 @@ bool ReportCFG::classof(const RejectReason *RR) {
 }
 
 //===----------------------------------------------------------------------===//
-// ReportNonBranchTerminator.
+// ReportInvalidTerminator.
 
-std::string ReportNonBranchTerminator::getMessage() const {
-  return ("Non branch instruction terminates BB: " + BB->getName()).str();
+std::string ReportInvalidTerminator::getMessage() const {
+  return ("Invalid instruction terminates BB: " + BB->getName()).str();
 }
 
-const DebugLoc &ReportNonBranchTerminator::getDebugLoc() const {
+const DebugLoc &ReportInvalidTerminator::getDebugLoc() const {
   return BB->getTerminator()->getDebugLoc();
 }
 
-bool ReportNonBranchTerminator::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkNonBranchTerminator;
-}
-
-//===----------------------------------------------------------------------===//
-// ReportCondition.
-
-std::string ReportCondition::getMessage() const {
-  return ("Not well structured condition at BB: " + BB->getName()).str();
-}
-
-const DebugLoc &ReportCondition::getDebugLoc() const {
-  return BB->getTerminator()->getDebugLoc();
-}
-
-bool ReportCondition::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkCondition;
+bool ReportInvalidTerminator::classof(const RejectReason *RR) {
+  return RR->getKind() == rrkInvalidTerminator;
 }
 
 //===----------------------------------------------------------------------===//
@@ -316,51 +287,6 @@ std::string ReportNonAffineAccess::getEndUserMessage() const {
   return "The array subscript of \"" + Name + "\" is not affine";
 }
 
-//===----------------------------------------------------------------------===//
-// ReportIndVar.
-
-ReportIndVar::ReportIndVar(const RejectReasonKind K) : RejectReason(K) {
-  ++BadIndVarForScop;
-}
-
-//===----------------------------------------------------------------------===//
-// ReportPhiNodeRefInRegion.
-
-ReportPhiNodeRefInRegion::ReportPhiNodeRefInRegion(Instruction *Inst)
-    : ReportIndVar(rrkPhiNodeRefInRegion), Inst(Inst) {}
-
-std::string ReportPhiNodeRefInRegion::getMessage() const {
-  return "SCEV of PHI node refers to SSA names in region: " + *Inst;
-}
-
-const DebugLoc &ReportPhiNodeRefInRegion::getDebugLoc() const {
-  return Inst->getDebugLoc();
-}
-
-bool ReportPhiNodeRefInRegion::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkPhiNodeRefInRegion;
-}
-
-//===----------------------------------------------------------------------===//
-// ReportIndEdge.
-
-ReportIndEdge::ReportIndEdge(BasicBlock *BB)
-    : RejectReason(rrkIndEdge), BB(BB) {
-  ++BadIndEdgeForScop;
-}
-
-std::string ReportIndEdge::getMessage() const {
-  return "Region has invalid entering edges!";
-}
-
-const DebugLoc &ReportIndEdge::getDebugLoc() const {
-  return BB->getTerminator()->getDebugLoc();
-}
-
-bool ReportIndEdge::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkIndEdge;
-}
-
 //===----------------------------------------------------------------------===//
 // ReportLoopBound.
 
@@ -410,6 +336,29 @@ bool ReportFuncCall::classof(const RejectReason *RR) {
   return RR->getKind() == rrkFuncCall;
 }
 
+//===----------------------------------------------------------------------===//
+// ReportNonSimpleMemoryAccess
+
+ReportNonSimpleMemoryAccess::ReportNonSimpleMemoryAccess(Instruction *Inst)
+    : ReportOther(rrkNonSimpleMemoryAccess), Inst(Inst) {}
+
+std::string ReportNonSimpleMemoryAccess::getMessage() const {
+  return "Non-simple memory access: " + *Inst;
+}
+
+const DebugLoc &ReportNonSimpleMemoryAccess::getDebugLoc() const {
+  return Inst->getDebugLoc();
+}
+
+std::string ReportNonSimpleMemoryAccess::getEndUserMessage() const {
+  return "Volatile memory accesses or memory accesses for atomic types "
+         "are not supported.";
+}
+
+bool ReportNonSimpleMemoryAccess::classof(const RejectReason *RR) {
+  return RR->getKind() == rrkNonSimpleMemoryAccess;
+}
+
 //===----------------------------------------------------------------------===//
 // ReportAlias.
 
@@ -468,21 +417,6 @@ bool ReportAlias::classof(const RejectReason *RR) {
   return RR->getKind() == rrkAlias;
 }
 
-//===----------------------------------------------------------------------===//
-// ReportSimpleLoop.
-
-ReportSimpleLoop::ReportSimpleLoop() : RejectReason(rrkSimpleLoop) {
-  ++BadSimpleLoopForScop;
-}
-
-std::string ReportSimpleLoop::getMessage() const {
-  return "Loop not in simplify form is invalid!";
-}
-
-bool ReportSimpleLoop::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkSimpleLoop;
-}
-
 //===----------------------------------------------------------------------===//
 // ReportOther.
 
@@ -549,24 +483,6 @@ bool ReportUnknownInst::classof(const RejectReason *RR) {
   return RR->getKind() == rrkUnknownInst;
 }
 
-//===----------------------------------------------------------------------===//
-// ReportPHIinExit.
-
-ReportPHIinExit::ReportPHIinExit(Instruction *Inst)
-    : ReportOther(rrkPHIinExit), Inst(Inst) {}
-
-std::string ReportPHIinExit::getMessage() const {
-  return "PHI node in exit BB";
-}
-
-const DebugLoc &ReportPHIinExit::getDebugLoc() const {
-  return Inst->getDebugLoc();
-}
-
-bool ReportPHIinExit::classof(const RejectReason *RR) {
-  return RR->getKind() == rrkPHIinExit;
-}
-
 //===----------------------------------------------------------------------===//
 // ReportEntry.
 ReportEntry::ReportEntry(BasicBlock *BB) : ReportOther(rrkEntry), BB(BB) {}
