@@ -95,51 +95,116 @@ RuntimeDebugBuilder::getGPUThreadIdentifiers(PollyIRBuilder &Builder) {
   return Identifiers;
 }
 
-void RuntimeDebugBuilder::createGPUVAPrinter(PollyIRBuilder &Builder,
-                                             ArrayRef<Value *> Values) {
+void RuntimeDebugBuilder::createPrinter(PollyIRBuilder &Builder, bool IsGPU,
+                                        ArrayRef<Value *> Values) {
+  if (IsGPU)
+    createGPUPrinterT(Builder, Values);
+  else
+    createCPUPrinterT(Builder, Values);
+}
+
+static std::tuple<std::string, std::vector<Value *>>
+prepareValuesForPrinting(PollyIRBuilder &Builder, ArrayRef<Value *> Values) {
+  std::string FormatString;
+  std::vector<Value *> ValuesToPrint;
+
+  for (auto Val : Values) {
+    Type *Ty = Val->getType();
+
+    if (Ty->isFloatingPointTy()) {
+      if (!Ty->isDoubleTy())
+        Val = Builder.CreateFPExt(Val, Builder.getDoubleTy());
+    } else if (Ty->isIntegerTy()) {
+      if (Ty->getIntegerBitWidth() < 64)
+        Val = Builder.CreateSExt(Val, Builder.getInt64Ty());
+      else
+        assert(Ty->getIntegerBitWidth() &&
+               "Integer types larger 64 bit not supported");
+    } else if (isa<PointerType>(Ty)) {
+      if (Ty->getPointerElementType() == Builder.getInt8Ty() &&
+          Ty->getPointerAddressSpace() == 4) {
+        Val = Builder.CreateGEP(Val, Builder.getInt64(0));
+      } else {
+        Val = Builder.CreatePtrToInt(Val, Builder.getInt64Ty());
+      }
+    } else {
+      llvm_unreachable("Unknown type");
+    }
+
+    Ty = Val->getType();
+
+    if (Ty->isFloatingPointTy())
+      FormatString += "%f";
+    else if (Ty->isIntegerTy())
+      FormatString += "%ld";
+    else
+      FormatString += "%s";
+
+    ValuesToPrint.push_back(Val);
+  }
+
+  return std::make_tuple(FormatString, ValuesToPrint);
+}
+
+void RuntimeDebugBuilder::createCPUPrinterT(PollyIRBuilder &Builder,
+                                            ArrayRef<Value *> Values) {
+
+  std::string FormatString;
+  std::vector<Value *> ValuesToPrint;
+
+  std::tie(FormatString, ValuesToPrint) =
+      prepareValuesForPrinting(Builder, Values);
+
+  createPrintF(Builder, FormatString, ValuesToPrint);
+  createFlush(Builder);
+}
+
+void RuntimeDebugBuilder::createGPUPrinterT(PollyIRBuilder &Builder,
+                                            ArrayRef<Value *> Values) {
   std::string str;
 
-  // Allocate print buffer (assuming 2*32 bit per element)
-  auto T = ArrayType::get(Builder.getInt32Ty(), Values.size() * 2);
-  Value *Data = new AllocaInst(
-      T, "polly.vprint.buffer",
-      Builder.GetInsertBlock()->getParent()->getEntryBlock().begin());
-
   auto *Zero = Builder.getInt64(0);
-  auto *DataPtr = Builder.CreateGEP(Data, {Zero, Zero});
 
   auto ToPrint = getGPUThreadIdentifiers(Builder);
 
   ToPrint.push_back(Builder.CreateGlobalStringPtr("\n  ", "", 4));
   ToPrint.insert(ToPrint.end(), Values.begin(), Values.end());
 
+  // Allocate print buffer (assuming 2*32 bit per element)
+  auto T = ArrayType::get(Builder.getInt32Ty(), ToPrint.size() * 2);
+  Value *Data = new AllocaInst(
+      T, "polly.vprint.buffer",
+      &Builder.GetInsertBlock()->getParent()->getEntryBlock().front());
+  auto *DataPtr = Builder.CreateGEP(Data, {Zero, Zero});
+
   int Offset = 0;
   for (auto Val : ToPrint) {
     auto Ptr = Builder.CreateGEP(DataPtr, Builder.getInt64(Offset));
     Type *Ty = Val->getType();
 
     if (Ty->isFloatingPointTy()) {
-      if (!Ty->isDoubleTy()) {
-        Ty = Builder.getDoubleTy();
-        Val = Builder.CreateFPExt(Val, Ty);
-      }
+      if (!Ty->isDoubleTy())
+        Val = Builder.CreateFPExt(Val, Builder.getDoubleTy());
     } else if (Ty->isIntegerTy()) {
-      auto Int64Bitwidth = Builder.getInt64Ty()->getIntegerBitWidth();
-      assert(Ty->getIntegerBitWidth() <= Int64Bitwidth);
-      if (Ty->getIntegerBitWidth() < Int64Bitwidth) {
-        Ty = Builder.getInt64Ty();
-        Val = Builder.CreateSExt(Val, Ty);
+      if (Ty->getIntegerBitWidth() < 64)
+        Val = Builder.CreateSExt(Val, Builder.getInt64Ty());
+      else
+        assert(Ty->getIntegerBitWidth() &&
+               "Integer types larger 64 bit not supported");
+    } else if (auto PtTy = dyn_cast<PointerType>(Ty)) {
+      if (PtTy->getAddressSpace() == 4) {
+        // Pointers in constant address space are printed as strings
+        Val = Builder.CreateGEP(Val, Builder.getInt64(0));
+        auto F = RuntimeDebugBuilder::getAddressSpaceCast(Builder, 4, 0);
+        Val = Builder.CreateCall(F, Val);
+      } else {
+        Val = Builder.CreatePtrToInt(Val, Builder.getInt64Ty());
       }
     } else {
-      // If it is not a number, it must be a string type.
-      Val = Builder.CreateGEP(Val, Builder.getInt64(0));
-      assert((Val->getType() == Builder.getInt8PtrTy(4)) &&
-             "Expected i8 ptr placed in constant address space");
-      auto F = RuntimeDebugBuilder::getAddressSpaceCast(Builder, 4, 0);
-      Val = Builder.CreateCall(F, Val);
-      Ty = Val->getType();
+      llvm_unreachable("Unknown type");
     }
 
+    Ty = Val->getType();
     Ptr = Builder.CreatePointerBitCastOrAddrSpaceCast(Ptr, Ty->getPointerTo(5));
     Builder.CreateAlignedStore(Val, Ptr, 4);
 
@@ -168,14 +233,24 @@ Function *RuntimeDebugBuilder::getPrintF(PollyIRBuilder &Builder) {
 
   if (!F) {
     GlobalValue::LinkageTypes Linkage = Function::ExternalLinkage;
-    FunctionType *Ty =
-        FunctionType::get(Builder.getInt32Ty(), Builder.getInt8PtrTy(), true);
+    FunctionType *Ty = FunctionType::get(Builder.getInt32Ty(), true);
     F = Function::Create(Ty, Linkage, Name, M);
   }
 
   return F;
 }
 
+void RuntimeDebugBuilder::createPrintF(PollyIRBuilder &Builder,
+                                       std::string Format,
+                                       ArrayRef<Value *> Values) {
+  Value *FormatString = Builder.CreateGlobalStringPtr(Format);
+  std::vector<Value *> Arguments;
+
+  Arguments.push_back(FormatString);
+  Arguments.insert(Arguments.end(), Values.begin(), Values.end());
+  Builder.CreateCall(getPrintF(Builder), Arguments);
+}
+
 void RuntimeDebugBuilder::createFlush(PollyIRBuilder &Builder) {
   Module *M = Builder.GetInsertBlock()->getParent()->getParent();
   const char *Name = "fflush";
@@ -196,30 +271,3 @@ void RuntimeDebugBuilder::createFlush(PollyIRBuilder &Builder) {
   // type to ensure that LLVM does not complain about mismatching types.
   Builder.CreateCall(F, Constant::getNullValue(F->arg_begin()->getType()));
 }
-
-void RuntimeDebugBuilder::createStrPrinter(PollyIRBuilder &Builder,
-                                           const std::string &String) {
-  Value *StringValue = Builder.CreateGlobalStringPtr(String);
-  Builder.CreateCall(getPrintF(Builder), StringValue);
-
-  createFlush(Builder);
-}
-
-void RuntimeDebugBuilder::createValuePrinter(PollyIRBuilder &Builder,
-                                             Value *V) {
-  const char *Format = nullptr;
-
-  Type *Ty = V->getType();
-  if (Ty->isIntegerTy())
-    Format = "%ld";
-  else if (Ty->isFloatingPointTy())
-    Format = "%lf";
-  else if (Ty->isPointerTy())
-    Format = "%p";
-
-  assert(Format && Ty->getPrimitiveSizeInBits() <= 64 && "Bad type to print.");
-
-  Value *FormatString = Builder.CreateGlobalStringPtr(Format);
-  Builder.CreateCall(getPrintF(Builder), {FormatString, V});
-  createFlush(Builder);
-}
