@@ -10,21 +10,6 @@
  * and Ecole Normale Superieure, 45 rue d'Ulm, 75230 Paris, France
  */
 
-#define xFN(TYPE,NAME) TYPE ## _ ## NAME
-#define FN(TYPE,NAME) xFN(TYPE,NAME)
-#define xS(TYPE,NAME) struct TYPE ## _ ## NAME
-#define S(TYPE,NAME) xS(TYPE,NAME)
-
-struct UNION {
-	int ref;
-#ifdef HAS_TYPE
-	enum isl_fold type;
-#endif
-	isl_space *space;
-
-	struct isl_hash_table	table;
-};
-
 __isl_give UNION *FN(UNION,cow)(__isl_take UNION *u);
 
 isl_ctx *FN(UNION,get_ctx)(__isl_keep UNION *u)
@@ -118,82 +103,6 @@ __isl_give UNION *FN(UNION,copy)(__isl_keep UNION *u)
 	return u;
 }
 
-/* Return the number of base expressions in "u".
- */
-int FN(FN(UNION,n),PARTS)(__isl_keep UNION *u)
-{
-	return u ? u->table.n : 0;
-}
-
-S(UNION,foreach_data)
-{
-	isl_stat (*fn)(__isl_take PART *part, void *user);
-	void *user;
-};
-
-static isl_stat FN(UNION,call_on_copy)(void **entry, void *user)
-{
-	PART *part = *entry;
-	S(UNION,foreach_data) *data = (S(UNION,foreach_data) *)user;
-
-	return data->fn(FN(PART,copy)(part), data->user);
-}
-
-isl_stat FN(FN(UNION,foreach),PARTS)(__isl_keep UNION *u,
-	isl_stat (*fn)(__isl_take PART *part, void *user), void *user)
-{
-	S(UNION,foreach_data) data = { fn, user };
-
-	if (!u)
-		return isl_stat_error;
-
-	return isl_hash_table_foreach(u->space->ctx, &u->table,
-				      &FN(UNION,call_on_copy), &data);
-}
-
-/* Is the space of "entry" equal to "space"?
- */
-static int FN(UNION,has_space)(const void *entry, const void *val)
-{
-	PART *part = (PART *)entry;
-	isl_space *space = (isl_space *) val;
-
-	return isl_space_is_equal(part->dim, space);
-}
-
-/* This function is not currently used by isl_aff.c.
- */
-static int FN(UNION,has_domain_space)(const void *entry, const void *val)
-	__attribute__ ((unused));
-
-/* Is the domain space of "entry" equal to "space"?
- */
-static int FN(UNION,has_domain_space)(const void *entry, const void *val)
-{
-	PART *part = (PART *)entry;
-	isl_space *space = (isl_space *) val;
-
-	if (isl_space_is_params(space))
-		return isl_space_is_set(part->dim);
-
-	return isl_space_tuple_is_equal(part->dim, isl_dim_in,
-					space, isl_dim_set);
-}
-
-/* Is the domain space of "entry" equal to the domain of "space"?
- */
-static int FN(UNION,has_same_domain_space)(const void *entry, const void *val)
-{
-	PART *part = (PART *)entry;
-	isl_space *space = (isl_space *) val;
-
-	if (isl_space_is_set(space))
-		return isl_space_is_set(part->dim);
-
-	return isl_space_tuple_is_equal(part->dim, isl_dim_in,
-					space, isl_dim_in);
-}
-
 /* Extract the element of "u" living in "space" (ignoring parameters).
  *
  * Return the ZERO element if "u" does not contain any element
@@ -202,7 +111,6 @@ static int FN(UNION,has_same_domain_space)(const void *entry, const void *val)
 __isl_give PART *FN(FN(UNION,extract),PARTS)(__isl_keep UNION *u,
 	__isl_take isl_space *space)
 {
-	uint32_t hash;
 	struct isl_hash_table_entry *entry;
 
 	if (!u || !space)
@@ -216,10 +124,10 @@ __isl_give PART *FN(FN(UNION,extract),PARTS)(__isl_keep UNION *u,
 			goto error;
 	}
 
-	hash = isl_space_get_hash(space);
-	entry = isl_hash_table_find(u->space->ctx, &u->table, hash,
-				    &FN(UNION,has_space), space, 0);
+	entry = FN(UNION,find_part_entry)(u, space, 0);
 	if (!entry)
+		goto error;
+	if (entry == isl_hash_table_entry_none)
 #ifdef HAS_TYPE
 		return FN(PART,ZERO)(space, u->type);
 #else
@@ -234,7 +142,8 @@ error:
 
 /* Add "part" to "u".
  * If "disjoint" is set, then "u" is not allowed to already have
- * a part that is defined on the same space as "part".
+ * a part that is defined over a domain that overlaps with the domain
+ * of "part".
  * Otherwise, compute the union sum of "part" and the part in "u"
  * defined on the same space.
  */
@@ -242,7 +151,6 @@ static __isl_give UNION *FN(UNION,add_part_generic)(__isl_take UNION *u,
 	__isl_take PART *part, int disjoint)
 {
 	int empty;
-	uint32_t hash;
 	struct isl_hash_table_entry *entry;
 
 	if (!part)
@@ -264,26 +172,18 @@ static __isl_give UNION *FN(UNION,add_part_generic)(__isl_take UNION *u,
 	if (!u)
 		goto error;
 
-	hash = isl_space_get_hash(part->dim);
-	entry = isl_hash_table_find(u->space->ctx, &u->table, hash,
-				    &FN(UNION,has_same_domain_space),
-				    part->dim, 1);
+	if (FN(UNION,check_disjoint_domain_other)(u, part) < 0)
+		goto error;
+	entry = FN(UNION,find_part_entry)(u, part->dim, 1);
 	if (!entry)
 		goto error;
 
 	if (!entry->data)
 		entry->data = part;
 	else {
-		PART *entry_part = entry->data;
-		if (disjoint)
-			isl_die(FN(UNION,get_ctx)(u), isl_error_invalid,
-				"additional part should live on separate "
-				"space", goto error);
-		if (!isl_space_tuple_is_equal(entry_part->dim, isl_dim_out,
-						part->dim, isl_dim_out))
-			isl_die(FN(UNION,get_ctx)(u), isl_error_invalid,
-				"union expression can only contain a single "
-				"expression over a given domain", goto error);
+		if (disjoint &&
+		    FN(UNION,check_disjoint_domain)(entry->data, part) < 0)
+			goto error;
 		entry->data = FN(PART,union_add_)(entry->data,
 						FN(PART,copy)(part));
 		if (!entry->data)
@@ -291,10 +191,8 @@ static __isl_give UNION *FN(UNION,add_part_generic)(__isl_take UNION *u,
 		empty = FN(PART,IS_ZERO)(part);
 		if (empty < 0)
 			goto error;
-		if (empty) {
-			FN(PART,free)(entry->data);
-			isl_hash_table_remove(u->space->ctx, &u->table, entry);
-		}
+		if (empty)
+			u = FN(UNION,remove_part_entry)(u, entry);
 		FN(PART,free)(part);
 	}
 
@@ -314,33 +212,138 @@ __isl_give UNION *FN(FN(UNION,add),PARTS)(__isl_take UNION *u,
 	return FN(UNION,add_part_generic)(u, part, 1);
 }
 
-static isl_stat FN(UNION,add_part)(__isl_take PART *part, void *user)
+#ifdef HAS_TYPE
+/* Allocate a UNION with the same type and the same size as "u" and
+ * with space "space".
+ */
+static __isl_give UNION *FN(UNION,alloc_same_size_on_space)(__isl_keep UNION *u,
+	__isl_take isl_space *space)
 {
-	UNION **u = (UNION **)user;
+	if (!u)
+		space = isl_space_free(space);
+	return FN(UNION,alloc)(space, u->type, u->table.n);
+}
+#else
+/* Allocate a UNION with the same size as "u" and with space "space".
+ */
+static __isl_give UNION *FN(UNION,alloc_same_size_on_space)(__isl_keep UNION *u,
+	__isl_take isl_space *space)
+{
+	if (!u)
+		space = isl_space_free(space);
+	return FN(UNION,alloc)(space, u->table.n);
+}
+#endif
 
-	*u = FN(FN(UNION,add),PARTS)(*u, part);
+/* Allocate a UNION with the same space, the same type (if any) and
+ * the same size as "u".
+ */
+static __isl_give UNION *FN(UNION,alloc_same_size)(__isl_keep UNION *u)
+{
+	return FN(UNION,alloc_same_size_on_space)(u, FN(UNION,get_space)(u));
+}
+
+/* Internal data structure for isl_union_*_transform_space.
+ * "fn' is applied to each entry in the input.
+ * "res" collects the results.
+ */
+S(UNION,transform_data)
+{
+	__isl_give PART *(*fn)(__isl_take PART *part, void *user);
+	void *user;
+
+	UNION *res;
+};
+
+/* Apply data->fn to "part" and add the result to data->res.
+ */
+static isl_stat FN(UNION,transform_entry)(__isl_take PART *part, void *user)
+{
+	S(UNION,transform_data) *data = (S(UNION,transform_data) *)user;
+
+	part = data->fn(part, data->user);
+	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
+	if (!data->res)
+		return isl_stat_error;
 
 	return isl_stat_ok;
 }
 
+/* Return a UNION living in "space" that is obtained by applying "fn"
+ * to each of the entries in "u".
+ */
+static __isl_give UNION *FN(UNION,transform_space)(__isl_take UNION *u,
+	isl_space *space,
+	__isl_give PART *(*fn)(__isl_take PART *part, void *user), void *user)
+{
+	S(UNION,transform_data) data = { fn, user };
+
+	data.res = FN(UNION,alloc_same_size_on_space)(u, space);
+	if (FN(FN(UNION,foreach),PARTS)(u,
+					&FN(UNION,transform_entry), &data) < 0)
+		data.res = FN(UNION,free)(data.res);
+	FN(UNION,free)(u);
+	return data.res;
+}
+
+/* Return a UNION that lives in the same space as "u" and that is obtained
+ * by applying "fn" to each of the entries in "u".
+ */
+static __isl_give UNION *FN(UNION,transform)(__isl_take UNION *u,
+	__isl_give PART *(*fn)(__isl_take PART *part, void *user), void *user)
+{
+	return FN(UNION,transform_space)(u, FN(UNION,get_space)(u), fn, user);
+}
+
+/* Apply data->fn to *part and store the result back into *part.
+ */
+static isl_stat FN(UNION,transform_inplace_entry)(void **part, void *user)
+{
+	S(UNION,transform_data) *data = (S(UNION,transform_data) *) user;
+
+	*part = data->fn(*part, data->user);
+	if (!*part)
+		return isl_stat_error;
+	return isl_stat_ok;
+}
+
+/* Update "u" by applying "fn" to each entry.
+ * This operation is assumed not to change the number of entries nor
+ * the spaces of the entries.
+ *
+ * If there is only one reference to "u", then change "u" inplace.
+ * Otherwise, create a new UNION from "u" and discard the original.
+ */
+static __isl_give UNION *FN(UNION,transform_inplace)(__isl_take UNION *u,
+	__isl_give PART *(*fn)(__isl_take PART *part, void *user), void *user)
+{
+	isl_bool single_ref;
+
+	single_ref = FN(UNION,has_single_reference)(u);
+	if (single_ref < 0)
+		return FN(UNION,free)(u);
+	if (single_ref) {
+		S(UNION,transform_data) data = { fn, user };
+		if (FN(UNION,foreach_inplace)(u,
+				&FN(UNION,transform_inplace_entry), &data) < 0)
+			return FN(UNION,free)(u);
+		return u;
+	}
+	return FN(UNION,transform)(u, fn, user);
+}
+
+/* An isl_union_*_transform callback for use in isl_union_*_dup
+ * that simply returns "part".
+ */
+static __isl_give PART *FN(UNION,copy_part)(__isl_take PART *part, void *user)
+{
+	return part;
+}
+
 __isl_give UNION *FN(UNION,dup)(__isl_keep UNION *u)
 {
-	UNION *dup;
-
-	if (!u)
-		return NULL;
-
-#ifdef HAS_TYPE
-	dup = FN(UNION,ZERO)(isl_space_copy(u->space), u->type);
-#else
-	dup = FN(UNION,ZERO)(isl_space_copy(u->space));
-#endif
-	if (FN(FN(UNION,foreach),PARTS)(u, &FN(UNION,add_part), &dup) < 0)
-		goto error;
-	return dup;
-error:
-	FN(UNION,free)(dup);
-	return NULL;
+	u = FN(UNION,copy)(u);
+	return FN(UNION,transform)(u, &FN(UNION,copy_part), NULL);
 }
 
 __isl_give UNION *FN(UNION,cow)(__isl_take UNION *u)
@@ -354,13 +357,6 @@ __isl_give UNION *FN(UNION,cow)(__isl_take UNION *u)
 	return FN(UNION,dup)(u);
 }
 
-static isl_stat FN(UNION,free_u_entry)(void **entry, void *user)
-{
-	PART *part = *entry;
-	FN(PART,free)(part);
-	return isl_stat_ok;
-}
-
 __isl_null UNION *FN(UNION,free)(__isl_take UNION *u)
 {
 	if (!u)
@@ -377,23 +373,13 @@ __isl_null UNION *FN(UNION,free)(__isl_take UNION *u)
 	return NULL;
 }
 
-S(UNION,align) {
-	isl_reordering *exp;
-	UNION *res;
-};
-
-static isl_stat FN(UNION,align_entry)(__isl_take PART *part, void *user)
+static __isl_give PART *FN(UNION,align_entry)(__isl_take PART *part, void *user)
 {
-	isl_reordering *exp;
-	S(UNION,align) *data = user;
+	isl_reordering *exp = user;
 
-	exp = isl_reordering_extend_space(isl_reordering_copy(data->exp),
+	exp = isl_reordering_extend_space(isl_reordering_copy(exp),
 				    FN(PART,get_domain_space)(part));
-
-	data->res = FN(FN(UNION,add),PARTS)(data->res,
-					    FN(PART,realign_domain)(part, exp));
-
-	return isl_stat_ok;
+	return FN(PART,realign_domain)(part, exp);
 }
 
 /* Reorder the parameters of "u" according to the given reordering.
@@ -401,23 +387,15 @@ static isl_stat FN(UNION,align_entry)(__isl_take PART *part, void *user)
 static __isl_give UNION *FN(UNION,realign_domain)(__isl_take UNION *u,
 	__isl_take isl_reordering *r)
 {
-	S(UNION,align) data = { NULL, NULL };
+	isl_space *space;
 
 	if (!u || !r)
 		goto error;
 
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(isl_space_copy(r->dim), u->type, u->table.n);
-#else
-	data.res = FN(UNION,alloc)(isl_space_copy(r->dim), u->table.n);
-#endif
-	data.exp = r;
-	if (FN(FN(UNION,foreach),PARTS)(u, &FN(UNION,align_entry), &data) < 0)
-		data.res = FN(UNION,free)(data.res);
-
-	isl_reordering_free(data.exp);
-	FN(UNION,free)(u);
-	return data.res;
+	space = isl_space_copy(r->dim);
+	u = FN(UNION,transform_space)(u, space, &FN(UNION,align_entry), r);
+	isl_reordering_free(r);
+	return u;
 error:
 	FN(UNION,free)(u);
 	isl_reordering_free(r);
@@ -522,36 +500,34 @@ S(UNION,match_bin_data) {
 	__isl_give PART *(*fn)(__isl_take PART *, __isl_take PART *);
 };
 
-/* Check if data->u2 has an element living in the same space as *entry.
+/* Check if data->u2 has an element living in the same space as "part".
  * If so, call data->fn on the two elements and add the result to
  * data->res.
  */
-static isl_stat FN(UNION,match_bin_entry)(void **entry, void *user)
+static isl_stat FN(UNION,match_bin_entry)(__isl_take PART *part, void *user)
 {
 	S(UNION,match_bin_data) *data = user;
-	uint32_t hash;
 	struct isl_hash_table_entry *entry2;
 	isl_space *space;
-	PART *part = *entry;
 	PART *part2;
 
 	space = FN(PART,get_space)(part);
-	hash = isl_space_get_hash(space);
-	entry2 = isl_hash_table_find(data->u2->space->ctx, &data->u2->table,
-				     hash, &FN(UNION,has_same_domain_space),
-				     space, 0);
+	entry2 = FN(UNION,find_part_entry)(data->u2, space, 0);
 	isl_space_free(space);
 	if (!entry2)
+		goto error;
+	if (entry2 == isl_hash_table_entry_none) {
+		FN(PART,free)(part);
 		return isl_stat_ok;
+	}
 
 	part2 = entry2->data;
 	if (!isl_space_tuple_is_equal(part->dim, isl_dim_out,
 					part2->dim, isl_dim_out))
 		isl_die(FN(UNION,get_ctx)(data->u2), isl_error_invalid,
 			"entries should have the same range space",
-			return isl_stat_error);
+			goto error);
 
-	part = FN(PART, copy)(part);
 	part = data->fn(part, FN(PART, copy)(entry2->data));
 
 	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
@@ -559,6 +535,9 @@ static isl_stat FN(UNION,match_bin_entry)(void **entry, void *user)
 		return isl_stat_error;
 
 	return isl_stat_ok;
+error:
+	FN(PART,free)(part);
+	return isl_stat_error;
 }
 
 /* This function is currently only used from isl_polynomial.c
@@ -584,13 +563,8 @@ static __isl_give UNION *FN(UNION,match_bin_op)(__isl_take UNION *u1,
 		goto error;
 
 	data.u2 = u2;
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(isl_space_copy(u1->space), u1->type,
-				    u1->table.n);
-#else
-	data.res = FN(UNION,alloc)(isl_space_copy(u1->space), u1->table.n);
-#endif
-	if (isl_hash_table_foreach(u1->space->ctx, &u1->table,
+	data.res = FN(UNION,alloc_same_size)(u1);
+	if (FN(FN(UNION,foreach),PARTS)(u1,
 				    &FN(UNION,match_bin_entry), &data) < 0)
 		goto error;
 
@@ -630,23 +604,15 @@ __isl_give UNION *FN(UNION,sub)(__isl_take UNION *u1, __isl_take UNION *u2)
 
 S(UNION,any_set_data) {
 	isl_set *set;
-	UNION *res;
 	__isl_give PW *(*fn)(__isl_take PW*, __isl_take isl_set*);
 };
 
-static isl_stat FN(UNION,any_set_entry)(void **entry, void *user)
+static __isl_give PART *FN(UNION,any_set_entry)(__isl_take PART *part,
+	void *user)
 {
 	S(UNION,any_set_data) *data = user;
-	PW *pw = *entry;
 
-	pw = FN(PW,copy)(pw);
-	pw = data->fn(pw, isl_set_copy(data->set));
-
-	data->res = FN(FN(UNION,add),PARTS)(data->res, pw);
-	if (!data->res)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return data->fn(part, isl_set_copy(data->set));
 }
 
 /* Update each element of "u" by calling "fn" on the element and "set".
@@ -655,7 +621,7 @@ static __isl_give UNION *FN(UNION,any_set_op)(__isl_take UNION *u,
 	__isl_take isl_set *set,
 	__isl_give PW *(*fn)(__isl_take PW*, __isl_take isl_set*))
 {
-	S(UNION,any_set_data) data = { NULL, NULL, fn };
+	S(UNION,any_set_data) data = { NULL, fn };
 
 	u = FN(UNION,align_params)(u, isl_set_get_space(set));
 	set = isl_set_align_params(set, FN(UNION,get_space)(u));
@@ -664,23 +630,12 @@ static __isl_give UNION *FN(UNION,any_set_op)(__isl_take UNION *u,
 		goto error;
 
 	data.set = set;
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->type,
-					u->table.n);
-#else
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->table.n);
-#endif
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
-				   &FN(UNION,any_set_entry), &data) < 0)
-		goto error;
-
-	FN(UNION,free)(u);
+	u = FN(UNION,transform)(u, &FN(UNION,any_set_entry), &data);
 	isl_set_free(set);
-	return data.res;
+	return u;
 error:
 	FN(UNION,free)(u);
 	isl_set_free(set);
-	FN(UNION,free)(data.res);
 	return NULL;
 }
 
@@ -716,29 +671,29 @@ static int FN(UNION,set_has_dim)(const void *entry, const void *val)
 }
 
 /* Find the set in data->uset that lives in the same space as the domain
- * of *entry, apply data->fn to *entry and this set (if any), and add
+ * of "part", apply data->fn to *entry and this set (if any), and add
  * the result to data->res.
  */
-static isl_stat FN(UNION,match_domain_entry)(void **entry, void *user)
+static isl_stat FN(UNION,match_domain_entry)(__isl_take PART *part, void *user)
 {
 	S(UNION,match_domain_data) *data = user;
 	uint32_t hash;
 	struct isl_hash_table_entry *entry2;
-	PW *pw = *entry;
 	isl_space *space;
 
-	space = FN(PW,get_domain_space)(pw);
+	space = FN(PART,get_domain_space)(part);
 	hash = isl_space_get_hash(space);
 	entry2 = isl_hash_table_find(data->uset->dim->ctx, &data->uset->table,
 				     hash, &FN(UNION,set_has_dim), space, 0);
 	isl_space_free(space);
-	if (!entry2)
+	if (!entry2) {
+		FN(PART,free)(part);
 		return isl_stat_ok;
+	}
 
-	pw = FN(PW,copy)(pw);
-	pw = data->fn(pw, isl_set_copy(entry2->data));
+	part = data->fn(part, isl_set_copy(entry2->data));
 
-	data->res = FN(FN(UNION,add),PARTS)(data->res, pw);
+	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
 	if (!data->res)
 		return isl_stat_error;
 
@@ -762,13 +717,8 @@ static __isl_give UNION *FN(UNION,match_domain_op)(__isl_take UNION *u,
 		goto error;
 
 	data.uset = uset;
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->type,
-					u->table.n);
-#else
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->table.n);
-#endif
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
+	data.res = FN(UNION,alloc_same_size)(u);
+	if (FN(FN(UNION,foreach),PARTS)(u,
 				   &FN(UNION,match_domain_entry), &data) < 0)
 		goto error;
 
@@ -795,31 +745,20 @@ __isl_give UNION *FN(UNION,intersect_domain)(__isl_take UNION *u,
 	return FN(UNION,match_domain_op)(u, uset, &FN(PW,intersect_domain));
 }
 
-/* Internal data structure for isl_union_*_subtract_domain.
- * uset is the set that needs to be removed from the domain.
- * res collects the results.
- */
-S(UNION,subtract_domain_data) {
-	isl_union_set *uset;
-	UNION *res;
-};
-
 /* Take the set (which may be empty) in data->uset that lives
  * in the same space as the domain of "pw", subtract it from the domain
- * of "pw" and add the result to data->res.
+ * of "part" and return the result.
  */
-static isl_stat FN(UNION,subtract_domain_entry)(__isl_take PW *pw, void *user)
+static __isl_give PART *FN(UNION,subtract_domain_entry)(__isl_take PART *part,
+	void *user)
 {
-	S(UNION,subtract_domain_data) *data = user;
+	isl_union_set *uset = user;
 	isl_space *space;
 	isl_set *set;
 
-	space = FN(PW,get_domain_space)(pw);
-	set = isl_union_set_extract_set(data->uset, space);
-	pw = FN(PW,subtract_domain)(pw, set);
-	data->res = FN(FN(UNION,add),PARTS)(data->res, pw);
-
-	return isl_stat_ok;
+	space = FN(PART,get_domain_space)(part);
+	set = isl_union_set_extract_set(uset, space);
+	return FN(PART,subtract_domain)(part, set);
 }
 
 /* Subtract "uset' from the domain of "u".
@@ -827,29 +766,9 @@ static isl_stat FN(UNION,subtract_domain_entry)(__isl_take PW *pw, void *user)
 __isl_give UNION *FN(UNION,subtract_domain)(__isl_take UNION *u,
 	__isl_take isl_union_set *uset)
 {
-	S(UNION,subtract_domain_data) data;
-
-	if (!u || !uset)
-		goto error;
-
-	data.uset = uset;
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->type,
-					u->table.n);
-#else
-	data.res = FN(UNION,alloc)(isl_space_copy(u->space), u->table.n);
-#endif
-	if (FN(FN(UNION,foreach),PARTS)(u,
-				&FN(UNION,subtract_domain_entry), &data) < 0)
-		data.res = FN(UNION,free)(data.res);
-
-	FN(UNION,free)(u);
+	u = FN(UNION,transform)(u, &FN(UNION,subtract_domain_entry), uset);
 	isl_union_set_free(uset);
-	return data.res;
-error:
-	FN(UNION,free)(u);
-	isl_union_set_free(uset);
-	return NULL;
+	return u;
 }
 
 __isl_give UNION *FN(UNION,gist)(__isl_take UNION *u,
@@ -860,59 +779,28 @@ __isl_give UNION *FN(UNION,gist)(__isl_take UNION *u,
 	return FN(UNION,match_domain_op)(u, uset, &FN(PW,gist));
 }
 
-#ifndef NO_EVAL
-__isl_give isl_val *FN(UNION,eval)(__isl_take UNION *u,
-	__isl_take isl_point *pnt)
-{
-	uint32_t hash;
-	struct isl_hash_table_entry *entry;
-	isl_space *space;
-	isl_val *v;
-
-	if (!u || !pnt)
-		goto error;
-
-	space = isl_space_copy(pnt->dim);
-	if (!space)
-		goto error;
-	hash = isl_space_get_hash(space);
-	entry = isl_hash_table_find(u->space->ctx, &u->table,
-				    hash, &FN(UNION,has_domain_space),
-				    space, 0);
-	isl_space_free(space);
-	if (!entry) {
-		v = isl_val_zero(isl_point_get_ctx(pnt));
-		isl_point_free(pnt);
-	} else {
-		v = FN(PART,eval)(FN(PART,copy)(entry->data), pnt);
-	}
-	FN(UNION,free)(u);
-	return v;
-error:
-	FN(UNION,free)(u);
-	isl_point_free(pnt);
-	return NULL;
-}
-#endif
-
+/* Coalesce an entry in a UNION.  Coalescing is performed in-place.
+ * Since the UNION may have several references, the entry is only
+ * replaced if the coalescing is successful.
+ */
 static isl_stat FN(UNION,coalesce_entry)(void **entry, void *user)
 {
-	PW **pw = (PW **)entry;
+	PART **part_p = (PART **) entry;
+	PART *part;
 
-	*pw = FN(PW,coalesce)(*pw);
-	if (!*pw)
+	part = FN(PART,copy)(*part_p);
+	part = FN(PW,coalesce)(part);
+	if (!part)
 		return isl_stat_error;
+	FN(PART,free)(*part_p);
+	*part_p = part;
 
 	return isl_stat_ok;
 }
 
 __isl_give UNION *FN(UNION,coalesce)(__isl_take UNION *u)
 {
-	if (!u)
-		return NULL;
-
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
-				   &FN(UNION,coalesce_entry), NULL) < 0)
+	if (FN(UNION,foreach_inplace)(u, &FN(UNION,coalesce_entry), NULL) < 0)
 		goto error;
 
 	return u;
@@ -947,16 +835,33 @@ error:
 	return NULL;
 }
 
-static isl_stat FN(UNION,mul_isl_int_entry)(void **entry, void *user)
+#ifdef HAS_TYPE
+/* Negate the type of "u".
+ */
+static __isl_give UNION *FN(UNION,negate_type)(__isl_take UNION *u)
+{
+	u = FN(UNION,cow)(u);
+	if (!u)
+		return NULL;
+	u->type = isl_fold_type_negate(u->type);
+	return u;
+}
+#else
+/* Negate the type of "u".
+ * Since "u" does not have a type, do nothing.
+ */
+static __isl_give UNION *FN(UNION,negate_type)(__isl_take UNION *u)
+{
+	return u;
+}
+#endif
+
+static __isl_give PART *FN(UNION,mul_isl_int_entry)(__isl_take PART *part,
+	void *user)
 {
-	PW **pw = (PW **)entry;
 	isl_int *v = user;
 
-	*pw = FN(PW,mul_isl_int)(*pw, *v);
-	if (!*pw)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PW,mul_isl_int)(part, *v);
 }
 
 __isl_give UNION *FN(UNION,mul_isl_int)(__isl_take UNION *u, isl_int v)
@@ -976,38 +881,21 @@ __isl_give UNION *FN(UNION,mul_isl_int)(__isl_take UNION *u, isl_int v)
 		return zero;
 	}
 
-	u = FN(UNION,cow)(u);
-	if (!u)
-		return NULL;
-
-#ifdef HAS_TYPE
+	u = FN(UNION,transform_inplace)(u, &FN(UNION,mul_isl_int_entry), &v);
 	if (isl_int_is_neg(v))
-		u->type = isl_fold_type_negate(u->type);
-#endif
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
-				    &FN(UNION,mul_isl_int_entry), &v) < 0)
-		goto error;
+		u = FN(UNION,negate_type)(u);
 
 	return u;
-error:
-	FN(UNION,free)(u);
-	return NULL;
 }
 
-/* Multiply *entry by the isl_val "user".
- *
- * Return 0 on success and -1 on error.
+/* Multiply "part" by the isl_val "user" and return the result.
  */
-static isl_stat FN(UNION,scale_val_entry)(void **entry, void *user)
+static __isl_give PART *FN(UNION,scale_val_entry)(__isl_take PART *part,
+	void *user)
 {
-	PW **pw = (PW **)entry;
 	isl_val *v = user;
 
-	*pw = FN(PW,scale_val)(*pw, isl_val_copy(v));
-	if (!*pw)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PART,scale_val)(part, isl_val_copy(v));
 }
 
 /* Multiply "u" by "v" and return the result.
@@ -1039,17 +927,9 @@ __isl_give UNION *FN(UNION,scale_val)(__isl_take UNION *u,
 		isl_die(isl_val_get_ctx(v), isl_error_invalid,
 			"expecting rational factor", goto error);
 
-	u = FN(UNION,cow)(u);
-	if (!u)
-		return NULL;
-
-#ifdef HAS_TYPE
+	u = FN(UNION,transform_inplace)(u, &FN(UNION,scale_val_entry), v);
 	if (isl_val_is_neg(v))
-		u->type = isl_fold_type_negate(u->type);
-#endif
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
-				    &FN(UNION,scale_val_entry), v) < 0)
-		goto error;
+		u = FN(UNION,negate_type)(u);
 
 	isl_val_free(v);
 	return u;
@@ -1059,20 +939,14 @@ error:
 	return NULL;
 }
 
-/* Divide *entry by the isl_val "user".
- *
- * Return 0 on success and -1 on error.
+/* Divide "part" by the isl_val "user" and return the result.
  */
-static isl_stat FN(UNION,scale_down_val_entry)(void **entry, void *user)
+static __isl_give PART *FN(UNION,scale_down_val_entry)(__isl_take PART *part,
+	void *user)
 {
-	PW **pw = (PW **)entry;
 	isl_val *v = user;
 
-	*pw = FN(PW,scale_down_val)(*pw, isl_val_copy(v));
-	if (!*pw)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PART,scale_down_val)(part, isl_val_copy(v));
 }
 
 /* Divide "u" by "v" and return the result.
@@ -1094,17 +968,9 @@ __isl_give UNION *FN(UNION,scale_down_val)(__isl_take UNION *u,
 		isl_die(isl_val_get_ctx(v), isl_error_invalid,
 			"cannot scale down by zero", goto error);
 
-	u = FN(UNION,cow)(u);
-	if (!u)
-		return NULL;
-
-#ifdef HAS_TYPE
+	u = FN(UNION,transform_inplace)(u, &FN(UNION,scale_down_val_entry), v);
 	if (isl_val_is_neg(v))
-		u->type = isl_fold_type_negate(u->type);
-#endif
-	if (isl_hash_table_foreach(FN(UNION,get_ctx)(u), &u->table,
-				    &FN(UNION,scale_down_val_entry), v) < 0)
-		goto error;
+		u = FN(UNION,negate_type)(u);
 
 	isl_val_free(v);
 	return u;
@@ -1123,16 +989,15 @@ S(UNION,plain_is_equal_data)
 static isl_stat FN(UNION,plain_is_equal_entry)(void **entry, void *user)
 {
 	S(UNION,plain_is_equal_data) *data = user;
-	uint32_t hash;
 	struct isl_hash_table_entry *entry2;
 	PW *pw = *entry;
 
-	hash = isl_space_get_hash(pw->dim);
-	entry2 = isl_hash_table_find(data->u2->space->ctx, &data->u2->table,
-				     hash, &FN(UNION,has_same_domain_space),
-				     pw->dim, 0);
-	if (!entry2) {
-		data->is_equal = isl_bool_false;
+	entry2 = FN(UNION,find_part_entry)(data->u2, pw->dim, 0);
+	if (!entry2 || entry2 == isl_hash_table_entry_none) {
+		if (!entry2)
+			data->is_equal = isl_bool_error;
+		else
+			data->is_equal = isl_bool_false;
 		return isl_stat_error;
 	}
 
@@ -1146,6 +1011,7 @@ static isl_stat FN(UNION,plain_is_equal_entry)(void **entry, void *user)
 isl_bool FN(UNION,plain_is_equal)(__isl_keep UNION *u1, __isl_keep UNION *u2)
 {
 	S(UNION,plain_is_equal_data) data = { NULL, isl_bool_true };
+	int n1, n2;
 
 	if (!u1 || !u2)
 		return isl_bool_error;
@@ -1153,6 +1019,12 @@ isl_bool FN(UNION,plain_is_equal)(__isl_keep UNION *u1, __isl_keep UNION *u2)
 		return isl_bool_true;
 	if (u1->table.n != u2->table.n)
 		return isl_bool_false;
+	n1 = FN(FN(UNION,n),PARTS)(u1);
+	n2 = FN(FN(UNION,n),PARTS)(u2);
+	if (n1 < 0 || n2 < 0)
+		return isl_bool_error;
+	if (n1 != n2)
+		return isl_bool_false;
 
 	u1 = FN(UNION,copy)(u1);
 	u2 = FN(UNION,copy)(u2);
@@ -1162,7 +1034,7 @@ isl_bool FN(UNION,plain_is_equal)(__isl_keep UNION *u1, __isl_keep UNION *u2)
 		goto error;
 
 	data.u2 = u2;
-	if (isl_hash_table_foreach(u1->space->ctx, &u1->table,
+	if (FN(UNION,foreach_inplace)(u1,
 			       &FN(UNION,plain_is_equal_entry), &data) < 0 &&
 	    data.is_equal)
 		goto error;
@@ -1177,61 +1049,23 @@ error:
 	return isl_bool_error;
 }
 
-#ifndef NO_NEG
-/* Replace *entry by its opposite.
- *
- * Return 0 on success and -1 on error.
- */
-static isl_stat FN(UNION,neg_entry)(void **entry, void *user)
-{
-	PW **pw = (PW **) entry;
-
-	*pw = FN(PW,neg)(*pw);
-
-	return *pw ? isl_stat_ok : isl_stat_error;
-}
-
-/* Return the opposite of "u".
- */
-__isl_give UNION *FN(UNION,neg)(__isl_take UNION *u)
-{
-	u = FN(UNION,cow)(u);
-	if (!u)
-		return NULL;
-
-	if (isl_hash_table_foreach(u->space->ctx, &u->table,
-				   &FN(UNION,neg_entry), NULL) < 0)
-		return FN(UNION,free)(u);
-
-	return u;
-}
-#endif
-
 /* Internal data structure for isl_union_*_drop_dims.
  * type, first and n are passed to isl_*_drop_dims.
- * res collects the results.
  */
 S(UNION,drop_dims_data) {
 	enum isl_dim_type type;
 	unsigned first;
 	unsigned n;
-
-	UNION *res;
 };
 
-/* Drop the parameters specified by "data" from "part" and
- * add the results to data->res.
+/* Drop the parameters specified by "data" from "part" and return the result.
  */
-static isl_stat FN(UNION,drop_dims_entry)(__isl_take PART *part, void *user)
+static __isl_give PART *FN(UNION,drop_dims_entry)(__isl_take PART *part,
+	void *user)
 {
 	S(UNION,drop_dims_data) *data = user;
 
-	part = FN(PART,drop_dims)(part, data->type, data->first, data->n);
-	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
-	if (!data->res)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PART,drop_dims)(part, data->type, data->first, data->n);
 }
 
 /* Drop the specified parameters from "u".
@@ -1253,45 +1087,28 @@ __isl_give UNION *FN(UNION,drop_dims)( __isl_take UNION *u,
 
 	space = FN(UNION,get_space)(u);
 	space = isl_space_drop_dims(space, type, first, n);
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(space, u->type, u->table.n);
-#else
-	data.res = FN(UNION,alloc)(space, u->table.n);
-#endif
-	if (FN(FN(UNION,foreach),PARTS)(u,
-					&FN(UNION,drop_dims_entry), &data) < 0)
-		data.res = FN(UNION,free)(data.res);
-
-	FN(UNION,free)(u);
-
-	return data.res;
+	return FN(UNION,transform_space)(u, space, &FN(UNION,drop_dims_entry),
+					&data);
 }
 
 /* Internal data structure for isl_union_*_set_dim_name.
  * pos is the position of the parameter that needs to be renamed.
  * s is the new name.
- * res collects the results.
  */
 S(UNION,set_dim_name_data) {
 	unsigned pos;
 	const char *s;
-
-	UNION *res;
 };
 
 /* Change the name of the parameter at position data->pos of "part" to data->s
- * and add the result to data->res.
+ * and return the result.
  */
-static isl_stat FN(UNION,set_dim_name_entry)(__isl_take PART *part, void *user)
+static __isl_give PART *FN(UNION,set_dim_name_entry)(__isl_take PART *part,
+	void *user)
 {
 	S(UNION,set_dim_name_data) *data = user;
 
-	part = FN(PART,set_dim_name)(part, isl_dim_param, data->pos, data->s);
-	data->res = FN(FN(UNION,add),PARTS)(data->res, part);
-	if (!data->res)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PART,set_dim_name)(part, isl_dim_param, data->pos, data->s);
 }
 
 /* Change the name of the parameter at position "pos" to "s".
@@ -1313,34 +1130,17 @@ __isl_give UNION *FN(UNION,set_dim_name)(__isl_take UNION *u,
 
 	space = FN(UNION,get_space)(u);
 	space = isl_space_set_dim_name(space, type, pos, s);
-#ifdef HAS_TYPE
-	data.res = FN(UNION,alloc)(space, u->type, u->table.n);
-#else
-	data.res = FN(UNION,alloc)(space, u->table.n);
-#endif
-
-	if (FN(FN(UNION,foreach),PARTS)(u,
-				    &FN(UNION,set_dim_name_entry), &data) < 0)
-		data.res = FN(UNION,free)(data.res);
-
-	FN(UNION,free)(u);
-
-	return data.res;
+	return FN(UNION,transform_space)(u, space,
+					&FN(UNION,set_dim_name_entry), &data);
 }
 
 /* Reset the user pointer on all identifiers of parameters and tuples
- * of the space of "part" and add the result to *res.
+ * of the space of "part" and return the result.
  */
-static isl_stat FN(UNION,reset_user_entry)(__isl_take PART *part, void *user)
+static __isl_give PART *FN(UNION,reset_user_entry)(__isl_take PART *part,
+	void *user)
 {
-	UNION **res = user;
-
-	part = FN(PART,reset_user)(part);
-	*res = FN(FN(UNION,add),PARTS)(*res, part);
-	if (!*res)
-		return isl_stat_error;
-
-	return isl_stat_ok;
+	return FN(PART,reset_user)(part);
 }
 
 /* Reset the user pointer on all identifiers of parameters and tuples
@@ -1349,23 +1149,9 @@ static isl_stat FN(UNION,reset_user_entry)(__isl_take PART *part, void *user)
 __isl_give UNION *FN(UNION,reset_user)(__isl_take UNION *u)
 {
 	isl_space *space;
-	UNION *res;
-
-	if (!u)
-		return NULL;
 
 	space = FN(UNION,get_space)(u);
 	space = isl_space_reset_user(space);
-#ifdef HAS_TYPE
-	res = FN(UNION,alloc)(space, u->type, u->table.n);
-#else
-	res = FN(UNION,alloc)(space, u->table.n);
-#endif
-	if (FN(FN(UNION,foreach),PARTS)(u,
-					&FN(UNION,reset_user_entry), &res) < 0)
-		res = FN(UNION,free)(res);
-
-	FN(UNION,free)(u);
-
-	return res;
+	return FN(UNION,transform_space)(u, space, &FN(UNION,reset_user_entry),
+					NULL);
 }
